import unittest

import numpy as np

from pystan import StanModel, stan
from pystan._compat import PY2
from pystan.tests.helper import get_model

# REF: rstan/tests/unitTests/runit.test.stanfit.R

class TestStanfit(unittest.TestCase):

    def test_init_zero_exception_inf_grad(self):
        code = """
        parameters {
            real x;
        }
        model {
            target += 1 / log(x);
        }
        """
        sm = StanModel(model_code=code)
        with self.assertRaises(RuntimeError):
            sm.sampling(init='0', iter=1, chains=1)

    def test_grad_log(self):
        y = np.array([0.70,  -0.16,  0.77, -1.37, -1.99,  1.35, 0.08,
                      0.02,  -1.48, -0.08,  0.34,  0.03, -0.42, 0.87,
                      -1.36,  1.43,  0.80, -0.48, -1.61, -1.27])

        code = '''
        data {
            int N;
            real y[N];
        }
        parameters {
            real mu;
            real<lower=0> sigma;
        }
        model {
            y ~ normal(mu, sigma);
        }'''

        def log_prob_fun(mu, log_sigma, adjust=True):
            sigma = np.exp(log_sigma)
            lp = -1 * np.sum((y - mu)**2) / (2 * (sigma**2)) - len(y) * np.log(sigma)
            if adjust:
                lp = lp + np.log(sigma)
            return lp

        def log_prob_grad_fun(mu, log_sigma, adjust=True):
            sigma = np.exp(log_sigma)
            g_lsigma = np.sum((y - mu)**2) * sigma**(-2) - len(y)
            if adjust:
                g_lsigma = g_lsigma + 1
            g_mu = np.sum(y - mu) * sigma**(-2)
            return (g_mu, g_lsigma)

        sm = get_model("normal_mu_sigma_model", code)
        sf = sm.sampling(data=dict(y=y, N=20), iter=200)
        mu = 0.1
        sigma = 2
        self.assertEqual(sf.log_prob(sf.unconstrain_pars(dict(mu=mu, sigma=sigma))),
                         log_prob_fun(mu, np.log(sigma)))
        self.assertEqual(sf.log_prob(sf.unconstrain_pars(dict(mu=mu, sigma=sigma)), False),
                         log_prob_fun(mu, np.log(sigma), adjust=False))
        g1 = sf.grad_log_prob(sf.unconstrain_pars(dict(mu=mu, sigma=sigma)), False)
        np.testing.assert_allclose(g1, log_prob_grad_fun(mu, np.log(sigma), adjust=False))

    def test_specify_args(self):
        y = (0.70,  -0.16,  0.77, -1.37, -1.99,  1.35, 0.08,
                0.02,  -1.48, -0.08,  0.34,  0.03, -0.42, 0.87,
                -1.36,  1.43,  0.80, -0.48, -1.61, -1.27)
        code = """
            data {
                int N;
                real y[N];
            }
            parameters {
                real mu;
                real<lower=0> sigma;
            }
            model {
                y ~ normal(mu, sigma);
            }"""
        stepsize0 = 0.15
        sm = get_model("normal_mu_sigma_model", code)
        sf = sm.sampling(data=dict(y=y, N=20), iter=200,
                         control=dict(adapt_engaged=False, stepsize=stepsize0))
        self.assertEqual(sf.get_sampler_params()[0]['stepsize__'][0], stepsize0)
        sf2 = stan(fit=sf, iter=20, algorithm='HMC', data=dict(y=y, N=20),
                    control=dict(adapt_engaged=False, stepsize=stepsize0))
        self.assertEqual(sf2.get_sampler_params()[0]['stepsize__'][0], stepsize0)
        sf3 = stan(fit=sf, iter=1, data=dict(y=y, N=20), init=0, chains=1)
        i_u = sf3.unconstrain_pars(sf3.get_inits()[0])
        np.testing.assert_equal(i_u, [0, 0])



import os
import pickle
import sys
import tempfile
import unittest

import pystan


class TestPickle(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.pickle_file = os.path.join(tempfile.mkdtemp(), 'stanmodel.pkl')
        cls.model_code = 'parameters {real y;} model {y ~ normal(0,1);}'

    def test_pickle_model(self):
        pickle_file = self.pickle_file
        model_code = self.model_code
        m = pystan.StanModel(model_code=model_code, model_name="normal2")
        module_name = m.module.__name__
        module_filename = m.module.__file__
        with open(pickle_file, 'wb') as f:
            pickle.dump(m, f)
        del m
        del sys.modules[module_name]

        with open(pickle_file, 'rb') as f:
            m = pickle.load(f)
        self.assertTrue(m.model_name.startswith("normal2"))
        self.assertIsNotNone(m.module)
        if not sys.platform.startswith('win'):
            # will fail on Windows
            self.assertNotEqual(module_filename, m.module.__file__)
        fit = m.sampling()
        y = fit.extract()['y']
        assert len(y) == 4000

    def test_pickle_fit(self):
        model_code = 'parameters {real y;} model {y ~ normal(0,1);}'

        sm = pystan.StanModel(model_code=model_code, model_name="normal1")

        # additional error checking
        fit = sm.sampling(iter=100)
        y = fit.extract()['y'].copy()
        self.assertIsNotNone(y)

        # pickle
        pickled_model = pickle.dumps(sm)
        module_name = sm.module.__name__
        del sm
        pickled_fit = pickle.dumps(fit)
        del fit

        # unload module
        if module_name in sys.modules:
            del(sys.modules[module_name])

        # load from file
        sm_from_pickle = pickle.loads(pickled_model)
        fit_from_pickle = pickle.loads(pickled_fit)
        self.assertIsNotNone(fit_from_pickle)
        self.assertTrue((fit_from_pickle.extract()['y'] == y).all())

    def test_pickle_model_and_reload(self):
        pickle_file = self.pickle_file
        pickle_file2 = os.path.join(tempfile.mkdtemp(), 'stanmodel.pkl')
        model_code = self.model_code
        model = pystan.StanModel(model_code=model_code, model_name="normal1")
        with open(pickle_file, 'wb') as f:
            pickle.dump(model, f)
        with open(pickle_file2, 'wb') as f:
            pickle.dump(model, f)

        del model

        with open(pickle_file, 'rb') as f:
            model_from_pickle = pickle.load(f)
        self.assertIsNotNone(model_from_pickle.sampling(iter=100).extract())
        with open(pickle_file2, 'rb') as f:
            model_from_pickle = pickle.load(f)
        self.assertIsNotNone(model_from_pickle.sampling(iter=100).extract())

    def test_model_unique_names(self):
        model_code = self.model_code
        model1 = pystan.StanModel(model_code=model_code, model_name="normal1")
        model2 = pystan.StanModel(model_code=model_code, model_name="normal1")
        self.assertNotEqual(model1.module_name, model2.module_name)



import distutils.errors
import sys
import os
import unittest

import pystan
from pystan._compat import PY2


class TestExtraCompileArgs(unittest.TestCase):

    def test_extra_compile_args(self):
        extra_compile_args = [
            '-O3',
            '-ftemplate-depth-1024',
            '-Wno-unused-function',
            '-Wno-uninitialized',
        ]
        if sys.platform.startswith("win"):
            extra_compile_args.extend([
                "-D_hypot=hypot",
                "-pthread",
                "-fexceptions"
            ])
        model_code = 'parameters {real y;} model {y ~ normal(0,1);}'
        model = pystan.StanModel(model_code=model_code, model_name="normal1",
                                 extra_compile_args=extra_compile_args)
        fit = model.sampling()
        extr = fit.extract()
        y_last, log_prob_last = extr['y'][-1], extr['lp__'][-1]
        self.assertEqual(fit.log_prob(y_last), log_prob_last)

    def test_extra_compile_args_failure(self):
        extra_compile_args = ['-non-existent-option']
        if sys.platform.startswith("win"):
            return
        model_code = 'parameters {real y;} model {y ~ normal(0,1);}'
        assertRaisesRegex = self.assertRaisesRegexp if PY2 else self.assertRaisesRegex
        with assertRaisesRegex(distutils.errors.CompileError, 'failed with exit status'):
            pystan.StanModel(model_code=model_code, model_name="normal1",
                             extra_compile_args=extra_compile_args)

    def test_threading_support(self):
        # Dont test with Windows with MinGW-w64 (GCC)
        if sys.platform.startswith("win"):
            return
        # Set up environmental variable
        os.environ['STAN_NUM_THREADS'] = "2"
        # Enable threading
        extra_compile_args = ['-pthread', '-DSTAN_THREADS']
        stan_code = """
        functions {
          vector bl_glm(vector mu_sigma, vector beta,
                        real[] x, int[] y) {
            vector[2] mu = mu_sigma[1:2];
            vector[2] sigma = mu_sigma[3:4];
            real lp = normal_lpdf(beta | mu, sigma);
            real ll = bernoulli_logit_lpmf(y | beta[1] + beta[2] * to_vector(x));
            return [lp + ll]';
          }
        }
        data {
          int<lower = 0> K;
          int<lower = 0> N;
          vector[N] x;
          int<lower = 0, upper = 1> y[N];
        }
        transformed data {
          int<lower = 0> J = N / K;
          real x_r[K, J];
          int<lower = 0, upper = 1> x_i[K, J];
          {
            int pos = 1;
            for (k in 1:K) {
              int end = pos + J - 1;
              x_r[k] = to_array_1d(x[pos:end]);
              x_i[k] = y[pos:end];
              pos += J;
            }
          }
        }
        parameters {
          vector[2] beta[K];
          vector[2] mu;
          vector<lower=0>[2] sigma;
        }
        model {
          mu ~ normal(0, 2);
          sigma ~ normal(0, 2);
          target += sum(map_rect(bl_glm, append_row(mu, sigma),
                                 beta, x_r, x_i));
        }
        """
        stan_data = dict(
            K = 4,
            N = 12,
            x = [1.204, -0.573, -1.35, -1.157,
                 -1.29, 0.515, 1.496, 0.918,
                 0.517, 1.092, -0.485, -2.157],
            y = [1, 1, 0, 0, 1, 1, 0, 0, 0, 1, 1, 1]
        )
        stan_model = pystan.StanModel(
            model_code=stan_code,
            extra_compile_args=extra_compile_args
        )
        for i in range(10):
            try:
                fit = stan_model.sampling(data=stan_data, chains=2, iter=200, n_jobs=1)
                self.assertIsNotNone(fit)
                fit2 = stan_model.sampling(data=stan_data, chains=2, iter=200, n_jobs=2)
                self.assertIsNotNone(fit2)
                draw = fit.extract(pars=fit.model_pars+['lp__'], permuted=False)
                lp = {key : values[-1, 0] for key, values in draw.items() if key == 'lp__'}['lp__']
                draw = {key : values[-1, 0] for key, values in draw.items() if key != 'lp__'}
                draw = fit.unconstrain_pars(draw)
                self.assertEqual(fit.log_prob(draw), lp)
                draw2 = fit2.extract(pars=fit2.model_pars+['lp__'], permuted=False)
                lp2 = {key : values[-1, 0] for key, values in draw2.items() if key == 'lp__'}['lp__']
                draw2 = {key : values[-1, 0] for key, values in draw2.items() if key != 'lp__'}
                draw2 = fit2.unconstrain_pars(draw2)
                self.assertEqual(fit2.log_prob(draw2), lp2)
                break
            except AssertionError:
                if i < 9:
                    continue
                else:
                    raise
